// SPDX-FileCopyrightText: 2022 Matteo Settenvini <matteo.settenvini@montecristosoftware.eu>
// SPDX-License-Identifier: GPL-3.0-or-later

#![allow(dead_code)]
include!(concat!(env!("OUT_DIR"), "/bindings.rs"));

use {
    crate::policy_checker::{MalcontentPolicyChecker, PolicyChecker as _},
    crate::utils::set_if_valid,
    libc::{hostent, size_t, socklen_t, AF_INET},
    nix::errno,
    nix::errno::Errno,
    std::ffi::CStr,
    std::os::raw::{c_char, c_int, c_void},
    std::ptr,
};

// See https://support.mozilla.org/en-US/kb/configuring-networks-disable-dns-over-https
const CANARY_HOSTNAME: &[u8] = b"use-application-dns.net\0";

// -------------- by host ---------------

#[no_mangle]
pub unsafe extern "C" fn _nss_malcontent_gethostbyname4_r(
    name: *const c_char,
    _pat: *mut *mut gaih_addrtuple,
    _buffer: *mut c_char,
    _buflen: size_t,
    errnop: *mut Errno,
    h_errnop: *mut HErrno,
    _ttlp: *mut i32,
) -> nss_status {
    let name = CStr::from_ptr(name);
    let policy_checker = MalcontentPolicyChecker::new(); // TODO: make LazySync, use RefCell for interior mutability
    match policy_checker.resolver(None) {
        Ok(None) => nss_status::NSS_STATUS_NOTFOUND,
        Ok(Some(resolver)) => {
            if name == CStr::from_bytes_with_nul_unchecked(CANARY_HOSTNAME) {
                set_if_valid(errnop, errno::from_i32(0));
                set_if_valid(h_errnop, HErrno::HostNotFound);
                return nss_status::NSS_STATUS_SUCCESS;
            }

            match resolver.lookup_ip(name.to_str().unwrap()) {
                Ok(_result) => {
                    // TODO
                    // https://sourceware.org/git/?p=glibc.git;a=blob;f=sysdeps/posix/getaddrinfo.c;h=fed2d3bf130b6fc88e5c6840804bbadb8fe9a98b;hb=4ab2ab03d4351914ee53248dc5aef4a8c88ff8b9
                    // https://chromium.googlesource.com/chromiumos/third_party/glibc/+/cvs/fedora-glibc-2_8_90-1/nss/nss_files/files-hosts.c
                    // https://github.com/systemd/systemd/blob/d5548eb618d426a3952746eb3a0e7d95d186ea7e/src/nss-resolve/nss-resolve.c

                    nss_status::NSS_STATUS_SUCCESS
                }
                Err(_) => nss_status::NSS_STATUS_UNAVAIL,
            }
        }
        Err(_err) => nss_status::NSS_STATUS_UNAVAIL,
    }
}

#[no_mangle]
pub unsafe extern "C" fn _nss_malcontent_gethostbyname3_r(
    _name: *const c_char,
    _af: c_int,
    _host: *mut hostent,
    _buffer: *mut c_char,
    _buflen: size_t,
    _errnop: *mut Errno,
    _h_errnop: *mut HErrno,
    _ttlp: *mut i32,
    _canonp: *mut *mut char,
) -> nss_status {
    todo!()
}

#[no_mangle]
pub unsafe extern "C" fn _nss_malcontent_gethostbyname2_r(
    name: *const c_char,
    af: c_int,
    host: *mut hostent,
    buffer: *mut c_char,
    buflen: size_t,
    errnop: *mut Errno,
    h_errnop: *mut HErrno,
) -> nss_status {
    _nss_malcontent_gethostbyname3_r(
        name,
        af,
        host,
        buffer,
        buflen,
        errnop,
        h_errnop,
        ptr::null_mut(),
        ptr::null_mut(),
    )
}

#[no_mangle]
pub unsafe extern "C" fn _nss_malcontent_gethostbyname_r(
    name: *const c_char,
    host: *mut hostent,
    buffer: *mut c_char,
    buflen: size_t,
    errnop: *mut Errno,
    h_errnop: *mut HErrno,
) -> nss_status {
    _nss_malcontent_gethostbyname3_r(
        name,
        AF_INET,
        host,
        buffer,
        buflen,
        errnop,
        h_errnop,
        ptr::null_mut(),
        ptr::null_mut(),
    )
}

// ----------------- by addr -----------------

#[no_mangle]
pub unsafe extern "C" fn _nss_malcontent_gethostbyaddr2_r(
    _addr: *const c_void,
    _len: socklen_t,
    _af: c_int,
    _host: *mut hostent,
    _buffer: *mut c_char,
    _buflen: size_t,
    _errnop: *mut Errno,
    _h_errnop: *mut HErrno,
    _ttlp: *mut i32,
) -> nss_status {
    todo!()
}

#[no_mangle]
pub unsafe extern "C" fn _nss_malcontent_gethostbyaddr_r(
    addr: *const c_void,
    len: socklen_t,
    af: c_int,
    host: *mut hostent,
    buffer: *mut c_char,
    buflen: size_t,
    errnop: *mut Errno,
    h_errnop: *mut HErrno,
) -> nss_status {
    _nss_malcontent_gethostbyaddr2_r(
        addr,
        len,
        af,
        host,
        buffer,
        buflen,
        errnop,
        h_errnop,
        ptr::null_mut(),
    )
}
